#!/usr/bin/env python3

# Generate visualizations for correlation dot products.
#
# Copyright © 2018-2024 C. Daniel Boschen
# Copyright © 2025 Remington Furman
#
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of the software demonstrated in code cells (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from numpy.random import randn
import scipy.signal as sig
import matplotlib.pyplot as plt

# Length of the correlation.
N = 50

# Time axis, signal is 1 second, sampled at 1 KSps.
t = np.linspace(0,1,N, endpoint = N)

# Modify this to add noise.
noise_mag = 0.0

# Modify this to add a phase offset.
phase_offset_rad = 0.0

# Input signal.
f = np.exp(1j*2*np.pi*3*t)
f = f + (noise_mag*randn(N) + noise_mag*1j*randn(N))/np.sqrt(2)

def mancorr(a,b):
    # Use shorter length if not same size.
    minsize = min(len(a),len(b))
    corr = np.dot(a[:minsize], np.conj(b[:minsize]))
    # Scale by length.
    return corr

def corr_vectors_pikchr(f,g,g_Hz,name):
    scale=10
    viewbox=(N+2)/scale
    out=""
    out+=f"line from (0,{-viewbox}) to (0,{viewbox})\n"
    out+=f"line from ({-viewbox},0) to ({viewbox},0)\n"
    out+="circlerad=0.1\n"
    out+="circle fill green at (0,0)\n"
    out+="dot invisible at last circle.c\n"
    for u,v in list(zip(f,g))[:-1]:
        prod = u*np.conj(v) / scale
        out+=f"arrow right {np.real(prod):0.4f} up {np.imag(prod):0.4f}\n"
    out+="circle fill red with .c at last arrow.end\n"

    out+=f'text "input: e^(j2π3t)" big big '
    out+=f'"test: {g_Hz:.2f} Hz" big big '
    out+=f'"window: {name}" big big at (1,1)\n'
    return out

def corr_visualization(name, window):
    f_windowed = f*window
    for g_Hz in np.arange(0.0, 10.01, 0.1):
        with open(f'{name}_{g_Hz:.2f}_Hz.pikchr', 'w') as file:
            # Correlate to g_Hz
            g = np.exp(1j*2*np.pi*g_Hz*t + 1j*phase_offset_rad)
            corr = mancorr(f_windowed, g)
            file.write(corr_vectors_pikchr(f_windowed,g,g_Hz,name))

def corr_plot(name, window):
    f_windowed = f*window

    # Sweep frequency from 0 to 10 Hz, 1000 samples.
    freqs = np.linspace(0, 10, 1000)
    corrs = [mancorr(f_windowed,
                     np.exp(1j*2*np.pi*f*t + 1j*phase_offset_rad))
             for f in freqs]
    
    plt.figure(figsize=(7,7))
    plt.subplot(2,1,1)
    plt.plot(freqs, np.abs(corrs))
    plt.title("Magnitude vs Frequency")
    plt.xlabel("Frequency (Hz)")
    plt.ylabel("Magnitude")
    plt.grid()
    ax = [0, 10, -10, N*1.2]
    plt.axis(ax)
    
    plt.subplot(2,1,2)
    plt.axis('square')
    plt.axis([-N,N,-N,N])
    plt.plot(np.real(corrs), np.imag(corrs))
    plt.title("Complex Plot vs Frequency")
    plt.xlabel("Real")
    plt.ylabel("Imaginary")
    plt.grid()
    plt.suptitle(r"Correlation to ($e^{j2\pi3t}$) vs Freq" + f" ({name})", fontsize=14)
    plt.tight_layout();
    plt.savefig(f'corr_plots_{name}.png')
    plt.close()

windows = {
    "boxcar": sig.windows.boxcar(N),
    "hamming": sig.windows.hamming(N),
    "blackman": sig.windows.blackman(N),
    "kaiser8": sig.windows.kaiser(N, 8),
    "kaiser12": sig.windows.kaiser(N, 12),
}

for window in windows:
    print(window)
    corr_visualization(window, windows[window])
    corr_plot(window, windows[window])

print("done")
